/* multiscatter_explicit.c -- Algorithm treating scattering orders explicitly

   Copyright (C) 2007 Robin Hogan <r.j.hogan@reading.ac.uk> 

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

/* For details of how to call the multiple scattering algorithms, see
   the comments in this header file: */
#include "multiscatter.h"

int ms_max_scattering_order = 4;

static int mse_num_gates_allocated = 0;
static int mse_num_gates_defined = 0;
static ms_real *mse_fs_fraction;
static ms_real *mse_theta2;
static ms_real *mse_rho2range2;

/* Allocate and initialize the intermediate arrays */
static
int
mse_init_intermediate_arrays(int num_gates)
{
  /* Macro for allocating or reallocating memory for a single array */
#define ALLOCATE(target, num_bytes) \
   if ((target = realloc(target, num_bytes)) == NULL) { \
      target = NULL; return MS_FAILURE; } 

  int i;
  if (mse_num_gates_allocated < num_gates) {
    /* Number of gates allocated is less than number requested (which
       may be zero if this function has not been called): (re)allocate
       memory. */
    int num_bytes = num_gates * sizeof(ms_real);
    mse_num_gates_allocated = 0; /* In case error occurs */
    ALLOCATE(mse_fs_fraction, num_bytes);
    ALLOCATE(mse_theta2, num_bytes);
    ALLOCATE(mse_rho2range2, num_bytes);
  }
  mse_num_gates_allocated = num_gates;
  /* Set each element of each array to zero. */
  for (i = -1; i <= num_gates; i++) {
    mse_fs_fraction[i] = 0.0;
    mse_theta2[i] = 0.0;
    mse_rho2range2[i] = 0.0;
  }
  mse_num_gates_defined = num_gates;
  return MS_SUCCESS;
}

/* Free the memory associated with the intermediate arrays */
static
void
mse_free_intermediate_arrays()
{
  if (mse_num_gates_allocated > 0) {
    free(mse_fs_fraction);
    free(mse_theta2);
    free(mse_rho2range2);
  }
  mse_num_gates_allocated = 0;
  mse_num_gates_defined = 0;
}

/* Function to be called recursively to calculate contributions from
   scattering up to order "norder" */
static
int
scatter(int n, int start_gate, int order, int norder, ms_real drange,
	ms_real energy, ms_real width2, ms_real zeta2, ms_real cov,
	ms_real *fs_fraction, ms_real *theta2, ms_real *rho2range2,
	ms_real *bscat_ratio_out) {
  ms_real start_factor = 0.5;
  int i;
  
  for (i = start_gate; i < n; i++) {
    ms_real range_diff = drange*(i-start_gate);
    ms_real width2i = width2 + 2.0*cov*range_diff
      + zeta2*range_diff*range_diff;
    /* Add backscatter contribution */
    ms_real bscat_contrib = energy*start_factor
      *(1.0-exp(-rho2range2[i]/width2i));
    bscat_ratio_out[i] += bscat_contrib;
    if (order == 2) {
      ms_bscat_double[i] += bscat_contrib;
    }
    else {
      ms_bscat_multi[i] += bscat_contrib;
    }
    if (ms_options & MS_OUTPUT_DISTRIBUTION && i == n-1) {
      msd_increment_distribution(order, i, energy*start_factor, width2i);
    }

    if (order < norder) {
      scatter(n, i, order+1, norder, drange,
	      energy*start_factor*fs_fraction[i], /* New energy */
	      width2i, /* New positional variance */
	      zeta2+theta2[i], /* New angular variance */
	      cov+zeta2*range_diff, /* New covariance */
	      fs_fraction, theta2, rho2range2, bscat_ratio_out);
    }
    start_factor = 1.0;
  }
  
  return MS_SUCCESS;
}

int
ms_set_max_scattering_order(int norder)
{
  if (norder > 1) {
    ms_max_scattering_order = norder;
    return MS_SUCCESS;
  }
  else {
    return MS_FAILURE;
  }
}

/* Use the photon variance-covariance approach to calculate the
   backscatter but treating each order of scattering explicitly. This
   is equivalent to Eloranta's formulation. */
int
multiscatter_explicit(int n,
		      ms_instrument instrument, ms_surface surface,
		      ms_real *range, ms_real *radius,
		      ms_real *ext, ms_real *ext_bscat_ratio,
		      ms_real *ext_air, ms_real *ssa_air,
		      ms_real *bscat_out)
{
  int i;
  ms_real drange = fabs(range[1]-range[0]);
  ms_real rho_transmitter2 = instrument.rho_transmitter
    *instrument.rho_transmitter;
  ms_real transmittance = 1.0, optical_depth = 0.0;
  ms_real bscat_ext_ratio_air = 3.0 / (8.0 * MS_PI);
  ms_real rho_factor = 1.0/(1-exp(-instrument.rho_receiver
				  *instrument.rho_receiver/
				  (instrument.rho_transmitter
				  *instrument.rho_transmitter)));

  /* Initialise arrays; note that the ms_bscat* variables defined in
     multiscatter_qsa.c are also initialised */
  if (mse_init_intermediate_arrays(n) == MS_FAILURE) {
    return MS_FAILURE;
  }
  if (ms_init_intermediate_arrays(n) == MS_FAILURE) {
    return MS_FAILURE;
  }

  /* If we are to output the lateral photon distribution then
     initialise the relevant 2D arrays */
  if (ms_options & MS_OUTPUT_DISTRIBUTION) {
    if (msd_init_intermediate_arrays(n) == MS_FAILURE) {
      return MS_FAILURE;
    }
  }

  /* Pre-compute some convenient variables */
  for (i = 0; i < n; i++) {
    ms_real theta = instrument.wavelength / (MS_PI*radius[i]);
    mse_theta2[i] = theta*theta;
    mse_fs_fraction[i] = ext[i]*drange;
    mse_rho2range2[i] = instrument.rho_receiver*instrument.rho_receiver
      * (range[i]-instrument.altitude)*(range[i]-instrument.altitude);
    bscat_out[i] = 0.0;
  }

  /* For double and higher-order scattering, call "scatter" which
     recursively calls itself for each subsequent order of scattering
     and each subsequent range gate */
  for (i = 0; i < n; i++) {
    ms_real range_abs = fabs(instrument.altitude-range[i]);
    scatter(n, i, 2, ms_max_scattering_order, drange, 
	    mse_fs_fraction[i],  /* Initial energy */
	    range_abs*range_abs*rho_transmitter2, /* Position variance */
	    rho_transmitter2+mse_theta2[i], /* Angular variance */
	    rho_transmitter2*range_abs, /* Covariance */
	    mse_fs_fraction, mse_theta2, mse_rho2range2,
	    bscat_out);
  }

  /* Work out the final backscatter */
  for (i = 0; i < n; i++) {
    ms_real od = ext[i]*drange; /* optical depth of layer */
    if (od > 0.0) {
      transmittance = exp(-optical_depth)*(1-exp(-od))/od;
      optical_depth += od;
    }
    else {
      transmittance = exp(-optical_depth);
    }
    /* The single-scattering backscatter is the sum of the
       contribution from cloud/aerosol and air, multiplied by the
       transmittance squared. Note that we are assuming that if the
       air molecules can scatter then they will behave as Rayleigh
       scatterers. */
    if (ext_air && ssa_air) {
      ms_bscat_single[i] = transmittance*transmittance
	*(ext[i]/ext_bscat_ratio[i] 
	  + ext_air[i]*ssa_air[i]*bscat_ext_ratio_air);
    }
    else {
      ms_bscat_single[i] = transmittance*transmittance*ext[i]
	/ext_bscat_ratio[i];
    }
    bscat_out[i] = ms_bscat_single[i]*(1+bscat_out[i]*rho_factor);
    /* Scale the double and triple-and-higher component
       appropriately */
    ms_bscat_double[i] *= ms_bscat_single[i]*rho_factor;
    ms_bscat_multi[i] *= ms_bscat_single[i]*rho_factor;
  }
  return MS_SUCCESS;
}

