/* multiscatter_fastqsa.c -- Fast QSA algorithm

   Copyright (C) 2007 Robin Hogan <r.j.hogan@reading.ac.uk> 

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

/* For details of how to call the multiple scattering algorithms, see
   the comments in this header file: */
#include "multiscatter.h"

#define SQRT1_2 0.70710678118654752440
#define SQRT_PI 1.77245385090552

/* Numerical recipes function for the complementary error function */
ms_real
erfc(ms_real x)
{
  ms_real t, z, ans;
  z = fabs(x);
  t = 1.0 / (1.0 + 0.5*z);
  ans = t * exp(-z*z - 1.26551223 + t*(1.00002368 + 
				       t*(0.37409196 + 
					  t*(0.09678418 + 
					     t*(-0.18628806 +
						t*(0.27886807 +
						   t*(-1.13520398 +
						      t*(1.48851587 +
							 t*(-0.82215223 +
							    t*0.17087277)))))))));
  return x >= 0.0 ? ans : 2.0-ans;
}

/* Use the photon variance-covariance approach to calculate the
   backscatter with O(N) efficiency */
int
multiscatter_fastqsa(int n,
		     ms_instrument instrument, ms_surface surface,
		     ms_real *range, ms_real *radius,
		     ms_real *ext, ms_real *ext_bscat_ratio,
		     ms_real *ext_air,
		     ms_real *droplet_fraction, ms_real *ice_fraction,
		     ms_real *bscat_out)
{
  int i;
  ms_real drange = fabs(range[1]-range[0]);
  ms_real drange2 = drange*drange;
  ms_real drange3 = drange2*drange;
  ms_real rho_transmitter2 = instrument.rho_transmitter
    *instrument.rho_transmitter;
  ms_real bscat_ext_ratio_air = 3.0 / (8.0 * MS_PI);
  ms_real rho_factor = 1.0/(1.0-exp(-instrument.rho_receiver
				  *instrument.rho_receiver/
				  (instrument.rho_transmitter
				  *instrument.rho_transmitter)));
  /* Properties at previous half-levels */
  ms_real midRange_prev = fabs(instrument.altitude-range[0])-drange*0.5;
  ms_real M_prev = 1.0;
  /* Initial properties of the total distribution */
  ms_real width2 = midRange_prev*midRange_prev*rho_transmitter2;
  ms_real zeta2 = rho_transmitter2;
  ms_real cov = midRange_prev*rho_transmitter2;
  ms_real trans = 1.0;
  /* Initial properties of the "reduced" total distribution */
  ms_real width2r = width2;
  ms_real zeta2r = zeta2;
  ms_real covr = cov;
  ms_real transr = trans;
  /* Initial properties of the unscattered distribution */
  ms_real transu = trans;
  ms_real transu_prev = transu;

  /* Main loop (note that this is the only one!) */
  for (i = 0; i < n; i++) {
    /* Variables at full levels */
    ms_real Theta = instrument.wavelength/(MS_PI*radius[i]);
    ms_real Theta2 = Theta*Theta;
    ms_real bscat_unattenuated = ext[i]/ext_bscat_ratio[i]
      + ext_air[i]*bscat_ext_ratio_air; 
    ms_real layer_od = 2.0*(ext[i]+ext_air[i])*drange;
    ms_real bscatu = bscat_unattenuated*transu_prev
	*(1.0-exp(-layer_od))/layer_od;
    /* Variables at half levels */
    ms_real Range = fabs(instrument.altitude-range[i]);
    ms_real midRange = Range+drange*0.5;
    ms_real midRange2 = midRange*midRange;
    ms_real width2a_max = midRange2
      *instrument.rho_receiver*instrument.rho_receiver;
    ms_real M = 1.0, transa, factora = 1.0;

    /* Properties of the unscattered distribution */
    ms_real width2u = midRange2*rho_transmitter2;
    ms_real covu = midRange*rho_transmitter2;
    
    /* Step the unscattered transmittance forward */
    transu = transu*exp(-layer_od);

    /* Step the total distribution at half-levels forward */
    width2 = width2 + zeta2*drange2
      + 2.0*cov*drange + ext[i]*Theta2*drange3*0.333333;
    cov = cov + zeta2*drange + ext[i]*Theta2*drange*0.5;
    zeta2 = zeta2 + ext[i]*drange*Theta2;
    trans = trans*exp(-(ext[i]+2.0*ext_air[i])*drange);
    
    /* Step the "reduced" total distribution at half-levels
       forward */
    width2r = width2r + zeta2r*drange2
      + 2.0*covr*drange + ext[i]*Theta2*drange3*0.333333;
    covr = covr + zeta2r*drange + ext[i]*Theta2*drange*0.5;
    zeta2r = zeta2r + ext[i]*drange*Theta2;
    transr = transr*exp(-(ext[i]+2.0*ext_air[i])*drange);

    /* Calculate the properties of the first Gaussian */
    transa = transr-transu;
    if (transa < 0.0) {
      transa = 0.0;
    }
    if (transa > transr*1.0e-12) {
      /* There has been some forward scattering */
      int is_anisotropic = droplet_fraction && ice_fraction 
	&& (droplet_fraction[i] > 0.0 || ice_fraction[i] > 0.0);
      ms_real width2a = (transr*width2r - transu*width2u)/transa;
      ms_real zeta2a = (transr*zeta2r - transu*rho_transmitter2)/transa;
      ms_real cova = (transr*covr - transu*covu)/transa;
      ms_real transb = trans - transr;
      if (width2a > width2a_max) {
	/* The first Gaussian is larger than the receiver FOV and
	   needs adjusting */
	ms_real correlation2a = cova*cova/(width2a*zeta2a);
	ms_real factor = width2a_max/width2a;
	//	fprintf(stderr, "ADJUSTING gate %d\n", i);
	width2a = width2a_max;
	transa = transa*factor;
	cova = cova*factor;
	zeta2a = zeta2a*(correlation2a*factor + 1-correlation2a);
	/* Recalculate "reduced" total distribution */
	transr = transu+transa;
	width2r = (transa*width2a + transu*width2u)/transr;
	zeta2r = (transa*zeta2a + transu*rho_transmitter2)/transr;
	covr = (transa*cova + transu*covu)/transr;
      }
      if (is_anisotropic) {
	/* The near-backscatter phase function is non-isotropic */
	ms_real gamma2a = zeta2a + width2a/midRange2 - 2.0*cova/midRange;
	ms_real droplet_factora = 1.0, ice_factora = 1.0;
	/* Parameterization for effect of the near-backscatter phase
	   function */
	if (droplet_fraction[i] > 0.0) {
	  droplet_factora = (1.0-MS_MIE_U1-MS_MIE_U2)
	    + MS_MIE_U1/(1.0 + gamma2a*MS_MIE_V1_SQD/Theta2)
	    + MS_MIE_U2/(1.0 + gamma2a*MS_MIE_V2_SQD/Theta2);
	}
	if (ice_fraction[i] > 0.0) {
	  ms_real g = sqrt(gamma2a)*0.5/MS_YANG_GAMMA0;
	  ice_factora = (1.0-MS_YANG_W)
	    + MS_YANG_W*(1.0-g*SQRT_PI*exp(g*g)*erfc(g));
	}
	/* The scaling factor for backscatter is the weighted sum of
	   the contributions from droplets, ice and air (the factor
	   for which is unity) */
	factora = (ext[i]*(droplet_factora*droplet_fraction[i]
			   + ice_factora*ice_fraction[i]
			   + 1.0-droplet_fraction[i]-ice_fraction[i])
		   +ext_air[i]) / (ext[i]+ext_air[i]);
      }

      if (transb > 0.0) {
	/* We have two Gaussians */
	ms_real width2b = (trans*width2 - transr*width2r)/transb;
	ms_real factorb = 1.0;;
	if (is_anisotropic) {
	  ms_real zeta2b = (trans*zeta2 - transr*zeta2r)/transb;
	  ms_real covb = (trans*cov - transr*covr)/transb;
	  ms_real covb_gamma = covb - width2b/midRange;
	  ms_real correlation2b_gamma = covb_gamma*covb_gamma/(width2b*zeta2b);
	  ms_real gamma2b = (zeta2b + width2b/midRange2 - 2.0*covb/midRange)
	    *(correlation2b_gamma*width2a_max/width2b + 1.0 - correlation2b_gamma);
	  ms_real droplet_factorb = 1.0, ice_factorb = 1.0;
	  /* Parameterization for effect of the near-backscatter phase
	     function */
	  if (droplet_fraction[i] > 0.0) {
	    droplet_factorb = (1.0-MS_MIE_U1-MS_MIE_U2)
	      + MS_MIE_U1/(1.0 + gamma2b*MS_MIE_V1_SQD/Theta2)
	      + MS_MIE_U2/(1.0 + gamma2b*MS_MIE_V2_SQD/Theta2);
	  }
	  if (ice_fraction[i] > 0.0) {
	    ms_real g = sqrt(gamma2b)*0.5/MS_YANG_GAMMA0;
	    ice_factorb = (1.0-MS_YANG_W)
	      + MS_YANG_W*(1.0-g*SQRT_PI*exp(g*g)*erfc(g));
	  }
	  /* The scaling factor for backscatter is the weighted sum of
	     the contributions from droplets, ice and air (the factor
	     for which is unity) */
	  factorb = (ext[i]*(droplet_factorb*droplet_fraction[i]
			     + ice_factorb*ice_fraction[i]
			     + 1.0-droplet_fraction[i]-ice_fraction[i])
		     +ext_air[i]) / (ext[i]+ext_air[i]);
	}
	M = 1.0 + ((1.0-exp(-width2a_max/width2a))*transa*factora
		   +(1.0-exp(-width2a_max/width2b))*transb*factorb)
	  * rho_factor / transu;
	fprintf(stderr, "%g %g\n", factora, factorb);
      }
      else {
	/* We have one Gaussian */
	M = 1.0 + rho_factor*factora*(1.0-exp(-width2a_max/width2a))*transa/transu;
      }
      bscat_out[i] = bscat_unattenuated*transu_prev
      	*(M_prev - M*exp(-layer_od))/(layer_od-log(M/M_prev));
      //      bscat_out[i] = bscatu*M;
      
      /*
      fprintf(stderr, "%d: %g total[%g %g %g %g] a[%g %g %g %g] M[%g %g]\n",
      	      i, bscat_out[i],trans, width2, zeta2, cov,
      	      transa, width2a, zeta2a, cova, M, M_prev);
      fprintf(stderr, "%d: width2a[%g %g] b[%g] %g %g %g\n", i, width2a, width2a_max,
	      transb, ext_air[i], bscat_unattenuated, transu_prev);
      */      
    }
    else if (layer_od > 0.0) {
      /* No forward scattering but some attenuation */
      bscat_out[i] = bscatu;
      /*
      fprintf(stderr, "%d: %g total[%g %g %g %g] a[%g] M[%g %g]\n",
      	      i, bscat_out[i],trans, width2, zeta2, cov,transa,M,M_prev);
      */
    }
    else {
      bscat_out[i] = 0.0;
    }
    /* Set variables for the previous half-level */
    midRange_prev = midRange;
    transu_prev = transu;
    M_prev = M;
  }

  return MS_SUCCESS;
}

